@@ -1,7 +1,7 @@
 /*-
- * See the file LICENSE for redistribution information.
+ * Copyright (c) 1996, 2020 Oracle and/or its affiliates.  All rights reserved.
  *
- * Copyright (c) 1996, 2013 Oracle and/or its affiliates.  All rights reserved.
+ * See the file LICENSE for license information.
  *
  * $Id$
  */
@@ -70,7 +70,7 @@ __memp_bhwrite(dbmp, hp, mfp, bhp, open_extents)
 			/* We may not be allowed to create backing files. */
 			if (mfp->no_backing_file) {
 				--dbmfp->ref;
-				return (EPERM);
+				return (USR_ERR(env, EPERM));
 			}
 
 			MUTEX_LOCK(env, dbmp->mutex);
@@ -99,7 +99,7 @@ __memp_bhwrite(dbmp, hp, mfp, bhp, open_extents)
 	 * It's the caller's choice if we're going to open extent files.
 	 */
 	if (!open_extents && F_ISSET(mfp, MP_EXTENT))
-		return (EPERM);
+		return (USR_ERR(env, EPERM));
 
 	/*
 	 * !!!
@@ -122,7 +122,7 @@ __memp_bhwrite(dbmp, hp, mfp, bhp, open_extents)
 	 * be marked dead.
 	 */
 	if (F_ISSET(mfp, MP_TEMP) || mfp->no_backing_file)
-		return (EPERM);
+		return (USR_ERR(env, EPERM));
 
 	/*
 	 * It's not a page from a file we've opened.  If the file requires
@@ -137,7 +137,7 @@ __memp_bhwrite(dbmp, hp, mfp, bhp, open_extents)
 				break;
 		MUTEX_UNLOCK(env, dbmp->mutex);
 		if (mpreg == NULL)
-			return (EPERM);
+			return (USR_ERR(env, EPERM));
 	}
 
 	/*
@@ -157,7 +157,7 @@ __memp_bhwrite(dbmp, hp, mfp, bhp, open_extents)
 	opened = 1;
 	if ((ret = __memp_fopen(dbmfp, mfp, NULL,
 	    NULL, DB_FLUSH | DB_DURABLE_UNKNOWN, 0, mfp->pagesize)) != 0) {
-	    	dbmfp->ref--;
+		dbmfp->ref--;
 		(void)__memp_fclose(dbmfp, 0);
 
 		/*
@@ -264,7 +264,7 @@ __memp_pgread(dbmfp, bhp, can_create)
 		 * how to handle the error.
 		 */
 		if (!can_create) {
-			ret = DB_PAGE_NOTFOUND;
+			ret = USR_ERR(env, DB_PAGE_NOTFOUND);
 			goto err;
 		}
 
@@ -356,7 +356,8 @@ __memp_pgwrite(env, dbmfp, hp, bhp)
 
 #ifdef DIAGNOSTIC
 	/*
-	 * Verify write-ahead logging semantics.
+	 * Verify write-ahead logging semantics, assuming the changes are
+	 * being logged, and the logs are durable.
 	 *
 	 * !!!
 	 * Two special cases.  There is a single field on the meta-data page,
@@ -374,7 +375,7 @@ __memp_pgwrite(env, dbmfp, hp, bhp)
 	 * the current end-of-log.
 	 */
 	if (LOGGING_ON(env) && !IS_NOT_LOGGED_LSN(LSN(bhp->buf)) &&
-	    !IS_CLIENT_PGRECOVER(env)) {
+	    !IS_CLIENT_PGRECOVER(env) && !F_ISSET(mfp, MP_NOT_DURABLE)) {
 		/*
 		 * There is a potential race here.  If we are in the midst of
 		 * switching log files, it's possible we could test against the
@@ -554,6 +555,9 @@ err:	__db_errx(env, DB_STR_A("3016",
  * __memp_bhfree --
  *	Free a bucket header and its referenced data.
  *
+ *	The hash bucket is unlocked before returning except when flags includes
+ *	BH_FREE_UNLOCKED -- or there was no hp passed in to begin with.
+ *
  * PUBLIC: int __memp_bhfree __P((DB_MPOOL *,
  * PUBLIC:	REGINFO *, MPOOLFILE *, DB_MPOOL_HASH *, BH *, u_int32_t));
  */
@@ -597,10 +601,13 @@ __memp_bhfree(dbmp, infop, mfp, hp, bhp, flags)
 	    (SH_CHAIN_NEXTP(bhp, vc, __bh)->td_off == bhp->td_off ||
 	    bhp->td_off == INVALID_ROFF ||
 	    IS_MAX_LSN(*VISIBLE_LSN(env, bhp)) ||
+	    F_ISSET(bhp, BH_UNREACHABLE) ||
 	    BH_OBSOLETE(bhp, hp->old_reader, vlsn))));
 
 	PERFMON3(env, mpool, evict, __memp_fns(dbmp, mfp), bhp->pgno, bhp);
-
+	if (FLD_ISSET(env->dbenv->verbose, DB_VERB_MVCC))
+		__db_msg(env, "bhfree pgno %lu roff %lx",
+		    (u_long)bhp->pgno, (u_long)R_OFFSET(dbmp->reginfo, bhp));
 	/*
 	 * Delete the buffer header from the hash bucket queue or the
 	 * version chain.
@@ -710,4 +717,3 @@ __memp_bh_clear_dirty(env, hp, bhp)
 		(void)atomic_dec(env, &hp->hash_page_dirty);
 	}
 }
-
