@@ -1,7 +1,7 @@
 /*-
- * See the file LICENSE for redistribution information.
+ * Copyright (c) 1996, 2020 Oracle and/or its affiliates.  All rights reserved.
  *
- * Copyright (c) 1996, 2013 Oracle and/or its affiliates.  All rights reserved.
+ * See the file LICENSE for license information.
  *
  * $Id$
  */
@@ -59,7 +59,8 @@ __memp_walk_files(env, mp, func, arg, countp, flags)
 			    mfp, arg, countp, flags)) != 0 && ret == 0)
 				ret = t_ret;
 			if (ret != 0 &&
-			    (!LF_ISSET(DB_STAT_MEMP_NOERROR) || ret == DB_BUFFER_SMALL))
+			    (!LF_ISSET(DB_STAT_MEMP_NOERROR) ||
+			    ret == DB_BUFFER_SMALL))
 				break;
 		}
 		MUTEX_UNLOCK(env, hp->mtx_hash);
@@ -98,9 +99,11 @@ __memp_discard_all_mpfs (env, mp)
 		while ((mfp = SH_TAILQ_FIRST(
 		    &hp->hash_bucket, __mpoolfile)) != NULL) {
 			MUTEX_LOCK(env, mfp->mutex);
-			if ((t_ret = __memp_mf_discard(dbmp, mfp, 1)) != 0 &&
-			    ret == 0)
-				ret = t_ret;
+			if ((t_ret = __memp_mf_discard(dbmp, mfp, 1)) != 0) {
+				if (ret == 0)
+					ret = t_ret;
+				break;
+			}
 		}
 		MUTEX_UNLOCK(env, hp->mtx_hash);
 	}
@@ -353,7 +356,7 @@ __memp_sync_int(env, dbmfp, trickle_max, flags, wrote_totalp, interruptedp)
 			 * we started.
 			 */
 #ifdef DIAGNOSTIC
-			if (SH_TAILQ_FIRST(&hp->hash_bucket, __bh) == NULL)
+			if (SH_TAILQ_EMPTY(&hp->hash_bucket))
 #else
 			if (atomic_read(&hp->hash_page_dirty) == 0)
 #endif
@@ -524,7 +527,7 @@ __memp_sync_int(env, dbmfp, trickle_max, flags, wrote_totalp, interruptedp)
 		}
 
 		/* Pin the buffer into memory. */
-		atomic_inc(env, &bhp->ref);
+		(void)atomic_inc(env, &bhp->ref);
 		MUTEX_UNLOCK(env, mutex);
 		MUTEX_READLOCK(env, bhp->mtx_buf);
 		DB_ASSERT(env, !F_ISSET(bhp, BH_EXCLUSIVE));
@@ -537,7 +540,7 @@ __memp_sync_int(env, dbmfp, trickle_max, flags, wrote_totalp, interruptedp)
 		 * dirty, so we can't just write it).
 		 */
 		if (SH_CHAIN_HASNEXT(bhp, vc)) {
-			atomic_dec(env, &bhp->ref);
+			(void)atomic_dec(env, &bhp->ref);
 			MUTEX_UNLOCK(env, bhp->mtx_buf);
 			continue;
 		}
@@ -569,7 +572,7 @@ __memp_sync_int(env, dbmfp, trickle_max, flags, wrote_totalp, interruptedp)
 			} else {
 				/* The buffer is being backed up, try again. */
 				if (t_ret == EAGAIN) {
-					atomic_dec(env, &bhp->ref);
+					(void)atomic_dec(env, &bhp->ref);
 					MUTEX_UNLOCK(env, bhp->mtx_buf);
 					continue;
 				}
@@ -588,7 +591,7 @@ __memp_sync_int(env, dbmfp, trickle_max, flags, wrote_totalp, interruptedp)
 
 		/* Discard our buffer reference. */
 		DB_ASSERT(env, atomic_read(&bhp->ref) > 0);
-		atomic_dec(env, &bhp->ref);
+		(void)atomic_dec(env, &bhp->ref);
 		MUTEX_UNLOCK(env, bhp->mtx_buf);
 
 		/* Check if the call has been interrupted. */
@@ -840,6 +843,7 @@ __memp_mf_sync(dbmp, mfp, locked)
 	MPOOLFILE *mfp;
 	int locked;
 {
+	APPNAME appname;
 	DB_FH *fhp;
 	DB_MPOOL_HASH *hp;
 	ENV *env;
@@ -849,6 +853,7 @@ __memp_mf_sync(dbmp, mfp, locked)
 
 	COMPQUIET(hp, NULL);
 	env = dbmp->env;
+	appname = DB_APP_DATA;
 
 	/*
 	 * We need to be holding the hash lock: we're using the path name
@@ -862,14 +867,24 @@ __memp_mf_sync(dbmp, mfp, locked)
 		MUTEX_LOCK(env, hp->mtx_hash);
 	}
 
-	if ((ret = __db_appname(env, DB_APP_DATA,
+mpsync:	if ((ret = __db_appname(env, appname,
 	    R_ADDR(dbmp->reginfo, mfp->path_off), NULL, &rpath)) == 0) {
 		if ((ret = __os_open(env, rpath, 0, 0, 0, &fhp)) == 0) {
 			ret = __os_fsync(env, fhp);
 			if ((t_ret =
 			    __os_closehandle(env, fhp)) != 0 && ret == 0)
 				ret = t_ret;
+		} else if (appname != DB_APP_BLOB) {
+			/* We may be syncing the blob meta db. */
+			__os_free(env, rpath);
+			appname = DB_APP_BLOB;
+			goto mpsync;
 		}
+		/* This usually means that it was removed outside of BDB. */
+		if (ret != 0)
+			__db_errx(env, DB_STR_A("3047",
+			    "__memp_mf_sync: Could not sync %s: %s", "%s %s"),
+			    rpath, db_strerror(ret));
 		__os_free(env, rpath);
 	}
 
@@ -971,13 +986,34 @@ __bhcmp(p1, p2)
  * __memp_purge_dead_files --
  *	Remove all dead files and their buffers from the mpool. The caller
  *	cannot hold any lock on the dead MPOOLFILE handles, their buffers
- *	or their hash buckets.
+ *	or their hash buckets. 
+ *	
+ *	Use this wrapper over __memp_purge_dead when the number of buffers
+ *	and dirty buffers is not of interest to the caller.
  *
  * PUBLIC: int __memp_purge_dead_files __P((ENV *));
  */
 int
 __memp_purge_dead_files(env)
 	ENV *env;
+{
+	return __memp_purge_dead(env, NULL, NULL);
+}
+
+/*
+ * __memp_purge_dead --
+ *	Remove all dead files and their buffers from the mpool. The caller
+ *	cannot hold any lock on the dead MPOOLFILE handles, their buffers
+ *	or their hash buckets.  Optionally return the total and dirty
+ *	buffer counts.
+ *
+ * PUBLIC: int __memp_purge_dead __P((ENV *, u_int32_t *, u_int32_t *));
+ */
+int
+__memp_purge_dead(env, totalp, dirtyp)
+	ENV *env;
+	u_int32_t *totalp;
+	u_int32_t *dirtyp;
 {
 	BH *bhp;
 	DB_MPOOL *dbmp;
@@ -985,7 +1021,7 @@ __memp_purge_dead_files(env)
 	REGINFO *infop;
 	MPOOL *c_mp, *mp;
 	MPOOLFILE *mfp;
-	u_int32_t i_cache;
+	u_int32_t i_cache, dirty, total;
 	int ret, t_ret, h_lock;
 
 	if (!MPOOL_ON(env))
@@ -994,31 +1030,34 @@ __memp_purge_dead_files(env)
 	dbmp = env->mp_handle;
 	mp = dbmp->reginfo[0].primary;
 	ret = t_ret = h_lock = 0;
+	dirty = total = 0;
 
 	/*
 	 * Walk each cache's list of buffers and free all buffers whose
 	 * MPOOLFILE is marked as dead.
 	 */
 	for (i_cache = 0; i_cache < mp->nreg; i_cache++) {
-		infop = &dbmp->reginfo[i_cache]; 
+		infop = &dbmp->reginfo[i_cache];
 		c_mp = infop->primary;
+		total += c_mp->pages;
 
 		hp = R_ADDR(infop, c_mp->htab);
 		hp_end = &hp[c_mp->htab_buckets];
 		for (; hp < hp_end; hp++) {
 			/* Skip empty buckets. */
-			if (SH_TAILQ_FIRST(&hp->hash_bucket, __bh) == NULL)
+			if (SH_TAILQ_EMPTY(&hp->hash_bucket))
 				continue;
 
-			/* 
+			dirty += (u_int32_t)atomic_read(&hp->hash_page_dirty);
+
+			/*
 			 * Search for a dead buffer. Other places that call
 			 * __memp_bhfree() acquire the buffer lock before the
 			 * hash bucket lock. Even though we acquire the two
 			 * locks in reverse order, we cannot deadlock here
 			 * because we don't block waiting for the locks.
 			 */
-			t_ret = MUTEX_TRYLOCK(env, hp->mtx_hash);
-			if (t_ret != 0) {
+			if ((t_ret = MUTEX_TRYLOCK(env, hp->mtx_hash)) != 0) {
 				__update_err_ret(t_ret, &ret);
 				continue;
 			}
@@ -1033,8 +1072,8 @@ __memp_purge_dead_files(env)
 					continue;
 
 				/* Found a dead buffer. Prepare to free it. */
-				t_ret = MUTEX_TRYLOCK(env, bhp->mtx_buf);
-				if (t_ret != 0) {
+				if ((t_ret =
+				    MUTEX_TRYLOCK(env, bhp->mtx_buf)) != 0) {
 					__update_err_ret(t_ret, &ret);
 					continue;
 				}
@@ -1075,6 +1114,11 @@ __memp_purge_dead_files(env)
 		}
 	}
 
+	if (dirtyp != NULL)
+		*dirtyp = dirty;
+	if (totalp != NULL)
+		*totalp = total;
+
 	return (ret);
 }
 
